import { useLayoutEffect, useRef, useState, useEffect, useMemo, createElement, useDebugValue, useCallback } from 'react';
import { unstable_runWithPriority, unstable_NormalPriority, unstable_UserBlockingPriority } from 'scheduler';
import { createContext, useContextUpdate, useContext, useContextSelector, useBridgeValue, BridgeProvider } from 'use-context-selector';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

const isSSR = typeof window === 'undefined' || /ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
const useIsoLayoutEffect = isSSR ? fn => fn() : useLayoutEffect;

const MAX_CHAIN_LENGTH = 10;
const SMALL_MAP_SIZE = 10;
const mCreate = () => [new Map()];
const mGet = (m, k) => {
  for (let i = 0; i < m.length; ++i) {
    const v = m[i].get(k);

    if (v) {
      return v;
    }
  }
};

const squash = m => {
  let i = m.length - 1;
  const dst = new Map(m[i]);

  while (i) {
    --i;
    m[i].forEach((v, k) => {
      dst.set(k, v);
    });
  }

  return [dst];
};

const mSet = (m, k, v) => {
  if (m[0].size < SMALL_MAP_SIZE) {
    return [new Map(m[0]).set(k, v), ...m.slice(1)];
  }

  if (m.length >= MAX_CHAIN_LENGTH) {
    m = squash(m);
  }

  return [new Map([[k, v]]), ...m];
};
const mDel = (m, k) => {
  if (m.length > 1) {
    m = squash(m);
  }

  const map = new Map(m[0]);
  map.delete(k);
  return [map];
};
const mKeys = m => {
  const keys = new Set();
  m.forEach(map => {
    for (const key of map.keys()) {
      keys.add(key);
    }
  });
  return keys;
};
const mMerge = (m, newM) => {
  m = [...newM, ...m];

  if (m.length > MAX_CHAIN_LENGTH) {
    m = squash(m);
  }

  return m;
};
const mToPrintable = (m, toPrintableK, toPrintableV) => {
  if (m.length > 1) {
    m = squash(m);
  }

  return new Map([...m[0].entries()].map(([k, v]) => [toPrintableK(k), toPrintableV(v)]));
};

const warningObject = new Proxy({}, {
  get() {
    throw new Error('Please use <Provider>');
  },

  apply() {
    throw new Error('Please use <Provider>');
  }

});

const addDependent = (dependentsMap, atom, dependent) => {
  let dependents = dependentsMap.get(atom);

  if (!dependents) {
    dependents = new Set();
    dependentsMap.set(atom, dependents);
  }

  dependents.add(dependent);
};

const deleteDependent = (dependentsMap, dependent) => {
  dependentsMap.forEach(dependents => {
    dependents.delete(dependent);
  });
};

const setDependencies = (dependentsMap, atom, dependencies) => {
  deleteDependent(dependentsMap, atom);
  dependencies.forEach(dependency => {
    addDependent(dependentsMap, dependency, atom);
  });
};

const listDependents = (dependentsMap, atom, excludeSelf) => {
  const dependents = new Set(dependentsMap.get(atom));

  if (excludeSelf) {
    dependents.delete(atom);
  }

  return dependents;
};

const initialState = mCreate();

const getAtomStateValue = (atom, state) => {
  const atomState = mGet(state, atom);

  if (!atomState) {
    console.warn('Atom state not found. Possibly a bug. Please file an issue with repro.');
    return atom.init;
  }

  return atomState.value;
};

const readAtomState = (atom, prevState, setState, dependentsMap, force) => {
  if (!force) {
    const atomState = mGet(prevState, atom);

    if (atomState) {
      return [atomState, prevState];
    }
  }

  let nextState = prevState;
  let error = undefined;
  let promise = undefined;
  let value = undefined;
  let dependencies = new Set();
  let isSync = true;

  try {
    const promiseOrValue = atom.read(a => {
      if (dependencies) {
        dependencies.add(a);
      } else {
        addDependent(dependentsMap, a, atom);
      }

      if (a !== atom) {
        const [aState, nextNextState] = readAtomState(a, nextState, setState, dependentsMap);

        if (isSync) {
          nextState = nextNextState;
        } else {
          // XXX is this really correct?
          setState(prev => mMerge(nextNextState, prev));
        }

        if (aState.readE) {
          throw aState.readE;
        }

        if (aState.readP) {
          throw aState.readP;
        }

        return aState.value;
      } // a === atom


      const aState = mGet(nextState, a);

      if (aState) {
        if (aState.readP) {
          throw aState.readP;
        }

        return aState.value;
      }

      return a.init; // this should not be undefined
    });

    if (promiseOrValue instanceof Promise) {
      promise = promiseOrValue.then(value => {
        setDependencies(dependentsMap, atom, dependencies);
        dependencies = null;
        setState(prev => {
          var _mGet;

          const prevPromise = (_mGet = mGet(prev, atom)) == null ? void 0 : _mGet.readP;

          if (prevPromise && prevPromise !== promise) {
            return prev;
          }

          return mSet(prev, atom, {
            value
          });
        });
      }).catch(e => {
        setState(prev => {
          var _mGet2;

          const prevPromise = (_mGet2 = mGet(prev, atom)) == null ? void 0 : _mGet2.readP;

          if (prevPromise && prevPromise !== promise) {
            return prev;
          }

          return mSet(prev, atom, {
            value: getAtomStateValue(atom, prev),
            readE: e instanceof Error ? e : new Error(e)
          });
        });
      });
    } else {
      setDependencies(dependentsMap, atom, dependencies);
      dependencies = null;
      value = promiseOrValue;
    }
  } catch (errorOrPromise) {
    if (errorOrPromise instanceof Promise) {
      promise = errorOrPromise;
    } else if (errorOrPromise instanceof Error) {
      error = errorOrPromise;
    } else {
      error = new Error(errorOrPromise);
    }
  }

  const nextAtomState = {
    readE: error,
    readP: promise,
    value: promise ? atom.init : value
  };
  nextState = mSet(nextState, atom, nextAtomState);
  isSync = false;
  return [nextAtomState, nextState];
};

const updateDependentsState = (atom, prevState, setState, dependentsMap) => {
  let nextState = prevState;
  listDependents(dependentsMap, atom, true).forEach(dependent => {
    if (typeof dependent === 'symbol') return;
    const [dependentState, nextNextState] = readAtomState(dependent, nextState, setState, dependentsMap, true);
    const promise = dependentState.readP;

    if (promise) {
      promise.then(() => {
        setState(prev => updateDependentsState(dependent, prev, setState, dependentsMap));
      });
      nextState = nextNextState;
    } else {
      nextState = updateDependentsState(dependent, nextNextState, setState, dependentsMap);
    }
  });
  return nextState;
};

const readAtom = (state, readingAtom, setState, dependentsMap, readPendingMap) => {
  let prevState = readPendingMap.get(state) || state;
  const [atomState, nextState] = readAtomState(readingAtom, prevState, setState, dependentsMap);

  if (nextState !== prevState) {
    readPendingMap.set(state, nextState);
  }

  return atomState;
};

const addAtom = (id, atom, dependentsMap) => {
  addDependent(dependentsMap, atom, id);
};

const delAtom = (id, setGcCount, dependentsMap) => {
  deleteDependent(dependentsMap, id);
  setGcCount(c => c + 1); // trigger re-render for gc
};

const gcAtom = (state, setState, dependentsMap) => {
  let nextState = state;
  let deleted;

  do {
    deleted = false;
    mKeys(nextState).forEach(atom => {
      var _dependentsMap$get;

      const isEmpty = ((_dependentsMap$get = dependentsMap.get(atom)) == null ? void 0 : _dependentsMap$get.size) === 0;

      if (isEmpty) {
        nextState = mDel(nextState, atom);
        dependentsMap.delete(atom);
        deleted = true;
      }
    });
  } while (deleted);

  if (nextState !== state) {
    setState(nextState);
  }
};

const writeAtom = (writingAtom, update, setState, dependentsMap, addWriteThunk) => {
  const pendingPromises = [];

  const writeAtomState = (prevState, atom, update) => {
    const prevAtomState = mGet(prevState, atom);

    if (prevAtomState && prevAtomState.writeP) {
      const promise = prevAtomState.writeP.then(() => {
        addWriteThunk(prev => writeAtomState(prev, atom, update));
      });
      pendingPromises.push(promise);
      return prevState;
    }

    let nextState = prevState;
    let isSync = true;

    try {
      const promiseOrVoid = atom.write(a => {
        if (process.env.NODE_ENV !== 'production') {
          const s = mGet(nextState, a);

          if (s && s.readP) {
            // TODO will try to detect this
            console.warn('Reading pending atom state in write operation. We need to detect this and fallback. Please file an issue for repro.', a);
          }
        }

        return getAtomStateValue(a, nextState);
      }, (a, v) => {
        if (a === atom) {
          const nextAtomState = {
            value: v
          };

          if (isSync) {
            nextState = mSet(nextState, a, nextAtomState);
            nextState = updateDependentsState(a, nextState, setState, dependentsMap);
          } else {
            setState(prev => updateDependentsState(a, mSet(prev, a, nextAtomState), setState, dependentsMap));
          }
        } else {
          if (isSync) {
            nextState = writeAtomState(nextState, a, v);
          } else {
            addWriteThunk(prev => writeAtomState(prev, a, v));
          }
        }
      }, update);

      if (promiseOrVoid instanceof Promise) {
        pendingPromises.push(promiseOrVoid);

        const nextAtomState = _extends({}, mGet(nextState, atom), {
          writeP: promiseOrVoid.then(() => {
            addWriteThunk(prev => mSet(prev, atom, _extends({}, mGet(prev, atom), {
              writeP: undefined
            })));
          })
        });

        nextState = mSet(nextState, atom, nextAtomState);
      }
    } catch (e) {
      if (pendingPromises.length) {
        pendingPromises.push(new Promise((_resolve, reject) => {
          reject(e);
        }));
      } else {
        throw e;
      }
    }

    isSync = false;
    return nextState;
  };

  addWriteThunk(prevState => writeAtomState(prevState, writingAtom, update));

  if (pendingPromises.length) {
    return new Promise((resolve, reject) => {
      const loop = () => {
        const len = pendingPromises.length;

        if (len === 0) {
          resolve();
        } else {
          Promise.all(pendingPromises).then(() => {
            pendingPromises.splice(0, len);
            loop();
          }).catch(reject);
        }
      };

      loop();
    });
  }
};

const runWriteThunk = (lastStateRef, pendingStateRef, setState, contextUpdate, writeThunkQueue) => {
  while (true) {
    if (!lastStateRef.current) {
      return;
    }

    if (writeThunkQueue.length === 0) {
      return;
    }

    const thunk = writeThunkQueue.shift();
    const prevState = pendingStateRef.current || lastStateRef.current;
    const nextState = thunk(prevState);

    if (nextState !== prevState) {
      pendingStateRef.current = nextState;
      unstable_runWithPriority(unstable_NormalPriority, () => {
        const pendingState = pendingStateRef.current;

        if (pendingState) {
          pendingStateRef.current = null;
          contextUpdate(() => {
            unstable_runWithPriority(unstable_UserBlockingPriority, () => {
              setState(pendingState);
            });
          });
        }
      });
    }
  }
};

const ActionsContext = createContext(warningObject);
const StateContext = createContext(warningObject);

const InnerProvider = ({
  r,
  children
}) => {
  const contextUpdate = useContextUpdate(StateContext);

  if (!r.current) {
    r.current = contextUpdate;
  }

  return children;
};

const Provider = ({
  children
}) => {
  const contextUpdateRef = useRef();
  const dependentsMapRef = useRef();

  if (!dependentsMapRef.current) {
    dependentsMapRef.current = new Map();
  }

  const readPendingMapRef = useRef();

  if (!readPendingMapRef.current) {
    readPendingMapRef.current = new WeakMap();
  }

  const [state, setStateOrig] = useState(initialState);
  const lastStateRef = useRef(null);
  const pendingStateRef = useRef(null);

  const setState = setStateAction => {
    lastStateRef.current = null;
    const pendingState = pendingStateRef.current;

    if (pendingState) {
      pendingStateRef.current = null;
      setStateOrig(pendingState);
    }

    setStateOrig(setStateAction);
  };

  useIsoLayoutEffect(() => {
    const readPendingMap = readPendingMapRef.current;
    const pendingState = readPendingMap.get(state);

    if (pendingState) {
      setState(pendingState);
      return;
    }

    lastStateRef.current = state;
  });
  const [gcCount, setGcCount] = useState(0); // to trigger gc

  useEffect(() => {
    gcAtom(state, setState, dependentsMapRef.current);
  }, [state, gcCount]);
  const writeThunkQueueRef = useRef([]);
  useEffect(() => {
    runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdateRef.current, writeThunkQueueRef.current);
  }, [state]);
  const actions = useMemo(() => ({
    add: (id, atom) => addAtom(id, atom, dependentsMapRef.current),
    del: id => delAtom(id, setGcCount, dependentsMapRef.current),
    read: (state, atom) => readAtom(state, atom, setState, dependentsMapRef.current, readPendingMapRef.current),
    write: (atom, update) => writeAtom(atom, update, setState, dependentsMapRef.current, thunk => {
      writeThunkQueueRef.current.push(thunk);
      runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdateRef.current, writeThunkQueueRef.current);
    })
  }), []);

  if (process.env.NODE_ENV !== 'production') {
    // eslint-disable-next-line react-hooks/rules-of-hooks
    useDebugState(state);
  }

  return createElement(ActionsContext.Provider, {
    value: actions
  }, createElement(StateContext.Provider, {
    value: state
  }, createElement(InnerProvider, {
    r: contextUpdateRef
  }, children)));
};

const useDebugState = state => {
  useDebugValue(state, s => mToPrintable(s, k => {
    var _k$debugLabel;

    return `${k.key}:${(_k$debugLabel = k.debugLabel) != null ? _k$debugLabel : '<no debugLabel>'}`;
  }, v => v.readE || v.readP || v.writeP || v.value));
};

let keyCount = 0; // global key count for all atoms

function atom(read, write) {
  const config = {
    key: ++keyCount
  };

  if (typeof read === 'function') {
    config.read = read;
  } else {
    config.init = read;

    config.read = get => get(config);

    config.write = (get, set, update) => {
      set(config, typeof update === 'function' ? update(get(config)) : update);
    };
  }

  if (write) {
    config.write = write;
  }

  return config;
}

const isWritable = atom => !!atom.write;

function useAtom(atom) {
  const actions = useContext(ActionsContext);
  const value = useContextSelector(StateContext, useCallback(state => {
    const atomState = actions.read(state, atom);

    if (atomState.readE) {
      throw atomState.readE; // read error
    }

    if (atomState.readP) {
      throw atomState.readP; // read promise
    }

    if (atomState.writeP) {
      throw atomState.writeP; // write promise
    }

    return atomState.value;
  }, [atom, actions]));
  useIsoLayoutEffect(() => {
    const id = Symbol();
    actions.add(id, atom);
    return () => {
      actions.del(id);
    };
  }, [actions, atom]);
  const setAtom = useCallback(update => {
    if (isWritable(atom)) {
      return actions.write(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [atom, actions]);
  useDebugValue(value);
  return [value, setAtom];
}

const useBridge = () => {
  const actions = useBridgeValue(ActionsContext);
  const state = useBridgeValue(StateContext);
  return useMemo(() => [actions, state], [actions, state]);
};
const Bridge = ({
  value,
  children
}) => {
  const [actions, state] = value;
  return createElement(BridgeProvider, {
    context: ActionsContext,
    value: actions
  }, createElement(BridgeProvider, {
    context: StateContext,
    value: state
  }, children));
};

export { Bridge, Provider, atom, useAtom, useBridge };
